﻿// -----------------------------------------------------------------------------------
// Nevatech, Inc. Copyright (C) 2016. All rights reserved.
// -----------------------------------------------------------------------------------

using Nevatech.Vsb.Repository;
using Nevatech.Vsb.Repository.Diagnostics;
using Nevatech.Vsb.Repository.Processing;
using Newtonsoft.Json;
using System;
using System.Net;
using System.ServiceModel.Channels;

namespace Nevatech.Vsb.Samples.CustomMessagePipelineComponent
{
    /// <summary>
    ///     Implements a custom message pipeline component that functions as a Circuit Breaker and limits the access
    ///     to the Physical Service if the number of errors during a period of time exceeds the limit.
    ///     In order to use this functionality, you need to place two instances of this custom component.
    ///     One will be in Inbound-Request pipeline section, and one in Inbound-Response.
    ///     
    ///     Custom Pipeline Component configuration:
    ///     Assembly: Nevatech.Vsb.Samples.CustomMessagePipelineComponent
    ///     Type: Nevatech.Vsb.Samples.CustomMessagePipelineComponent.CircuitBreakerComponent
    /// </summary>
    public sealed class CircuitBreakerComponent : IMessageProcessor
    {
        #region Configuration class

        /// <summary>
        ///     Represents the component configuration.
        ///     Default value will be used if the property
        ///     is not present in a JSON configuration string.
        /// </summary>
        public class Configuration
        {
            /// <summary>
            ///     Gets or sets the maximum allowed number of Errors that will be counted during Error Timer.
            /// </summary>
            public int ErrorLimit { get; set; } = 5;

            /// <summary>
            ///     Gets or sets the Error Timer (seconds). If Error Limit is reached within the Error Timer period,
            ///     the Circuit Breaker will switch to Open state.
            /// </summary>
            public int ErrorTimer { get; set; } = 10;

            /// <summary>
            ///     Gets or sets the Open Timer (seconds). It is a period of time that the Circuit Breaker will
            ///     spend in Open state. All requests will be blocked during this time.
            /// </summary>
            public int OpenTimer { get; set; } = 10;

            /// <summary>
            ///     Gets or sets the HTTP Status Code of the message that the Circuit Breaker
            ///     will return when the message is blocked during Open Timer period.
            /// </summary>
            public int StatusCode { get; set; } = 503;

            /// <summary>
            ///     Gets or sets the Content-Type header of the message that the Circuit Breaker will
            ///     return when the message is blocked. Context property tokens are supported.
            /// </summary>
            public string ContentType { get; set; } = "application/json";

            /// <summary>
            ///     Gets or sets the response body of the message that the Circuit Breaker will
            ///     return when the message is blocked. Context property tokens are supported.
            /// </summary>
            public string Body { get; set; } = "Circuit Breaker is open - Request is blocked";

        }

        #endregion


        #region Import Configuration
        private Configuration _configuration;

        /// <summary>
        ///     Imports component configuration.
        /// </summary>
        /// <param name="configuration">
        ///     JSON string that specifies the Circuit Breaker parameters.
        ///     Circuit Breaker response is sent when the request is blocked.
        ///     This response is configured in Inbound-Request pipeline section.
        ///     Inbound-Request template:
        /// <![CDATA[
        ///     {
        ///         "StatusCode": 503,
        ///         "ContentType": "application/json",
        ///         "Body": "Circuit Breaker is open - Request is blocked"
        ///     }
        /// ]]>
        ///     Circuit Breaker parameters are configured in Inbound-Response pipeline section.
        ///     Inbound-Response template:
        /// <![CDATA[
        ///     {
        ///         "ErrorLimit": 5,
        ///         "ErrorTimer": 10,
        ///         "OpenTimer": 10
        ///     }
        /// ]]>
        /// </param>
        public void ImportConfiguration(string configuration)
        {
            _configuration = !String.IsNullOrWhiteSpace(configuration) ?
                JsonConvert.DeserializeObject<Configuration>(configuration) :
                new Configuration();
        }

        #endregion        

        /// <summary>
        ///     CircuitBreaker class is used to control the processing flow.
        /// </summary>
        private class CircuitBreaker
        {
            // Names of Circuit Breaker states.            
            public const string Closed = "Closed";
            public const string HalfOpen = "Half-Open";
            public const string Open = "Open";
            
            // Names of Cache Properties.            
            public const string State = "State";
            public const string ErrorTimer = "Error Timer";
            public const string OpenTimer = "Open Timer";

            /// <summary>
            ///    Expiration that will be used when initializing a new Error Timer Cache Property.
            /// </summary>
            private readonly TimeSpan errorTimerExpiration;

            /// <summary>
            ///     Expiration that will be used when initializing a new Open Timer Cache Property.
            /// </summary>
            private readonly TimeSpan openTimerExpiration;

            /// <summary>
            ///     HTTP Status Code of the message that the Circuit Breaker will return when the message is blocked.
            /// </summary>
            private readonly HttpStatusCode responseStatusCode;

            /// <summary>
            ///     Content-Type header of the message that the Circuit Breaker will return when the message is blocked.
            /// </summary>
            private readonly string responseContentType;

            /// <summary>
            ///     Body of the message that the Circuit Breaker will return when the message is blocked.
            /// </summary>
            private readonly string responseBody;

            /// <summary>
            ///     Gets the Circuit Breaker Error Limit.
            /// </summary>
            public int ErrorLimit { get; }

            /// <summary>
            ///     Message Pipeline Context.
            /// </summary>
            private MessagePipelineContext circuitBreakerContext;

            /// <summary>
            ///     CircuitBreaker Constructor.
            /// </summary>
            public CircuitBreaker(Configuration configuration, MessagePipelineContext inputContext)
            {
                ErrorLimit = configuration.ErrorLimit;

                errorTimerExpiration = TimeSpan.FromSeconds(configuration.ErrorTimer);
                openTimerExpiration = TimeSpan.FromSeconds(configuration.OpenTimer);

                circuitBreakerContext = inputContext;

                responseStatusCode = (HttpStatusCode)configuration.StatusCode;
                responseContentType = circuitBreakerContext.ReplaceContextPropertyTokens(configuration.ContentType);
                responseBody = circuitBreakerContext.ReplaceContextPropertyTokens(configuration.Body);                
            }

            /// <summary>
            ///     Any response HTTP status code outside of the 100-399 range will increase the Error Counter by 1.
            /// </summary>
            public bool IsExpectedResponse(int httpStatusCode)
            {
                return 100 <= httpStatusCode && httpStatusCode <= 399;
            }

            /// <summary>
            ///     Returns current Circuit Breaker data by capturing it from a Cache Property.
            /// </summary>
            public string ReturnCurrentData(string propName)
            {
                return CacheProperty.Get(circuitBreakerContext.Configuration.Key, propName);
            }

            /// <summary>
            ///     Returns current Circuit Breaker Counter data by capturing it from a Cache Property.
            /// </summary>
            public int ReturnCurrentCounterData()
            {                
                return CacheProperty.GetOrSet<int>(circuitBreakerContext.Configuration.Key, ErrorTimer);
            }

            /// <summary>
            ///     Switch Circuit Breaker State.
            /// </summary>
            public void SwitchToState(string newState)
            {
                CacheProperty.Set(circuitBreakerContext.Configuration.Key, State, newState, null, true);

                if (newState == Open)
                {
                    CacheProperty.Set(circuitBreakerContext.Configuration.Key, OpenTimer, "Open Timer Value", openTimerExpiration, false);
                    CacheProperty.Remove(circuitBreakerContext.Configuration.Key, ErrorTimer);
                }
            }

            /// <summary>
            ///     Configure Circuit Breaker response message.
            ///     It will be sent when the request is blocked.
            /// </summary>
            public void ConfigureResponse()
            {
                HttpResponseMessageProperty httpResponse = circuitBreakerContext.Message.GetHttpResponseMessageProperty(true);
                httpResponse.Headers[System.Net.HttpResponseHeader.ContentType] = responseContentType;
                circuitBreakerContext.SetMessageBodyContent(responseBody);
                httpResponse.StatusCode = responseStatusCode;
            }            

            /// <summary>
            ///     Increase Error Current by 1.
            /// </summary>
            public void IncreaseErrorCounter()
            {                
                CacheProperty.GetOrSet<int>(circuitBreakerContext.Configuration.Key, ErrorTimer,
                    () => 1, (count) => count + 1, errorTimerExpiration, false);
            }

            /// <summary>
            ///     Returns true if Error Limit is reached.
            /// </summary>
            public bool IsErrorLimitReached()
            {
                return ReturnCurrentCounterData() >= ErrorLimit;
            }
        }

        /// <summary>
        ///     Process Message using Circuit Breaker that limits access to the Physical Service if a
        ///     number of errors during a period of time exceeds the limit.
        /// </summary>
        /// <param name="context">
        ///     Provides the message processing context.
        /// </param>
        /// <returns>
        ///     Value indicating if message should continue on its way to the recipient or
        ///     must be stopped and returned to the sender.
        /// </returns>
        public MessagePipelineResult ProcessMessage(MessagePipelineContext context)
        {
            if (context == null)
            {
                throw new ArgumentNullException("context");
            }

            CircuitBreaker circuitBreaker = new CircuitBreaker(_configuration, context);            

            TraceMark traceMark = context.TraceWriter.WriteStart("CIRCUIT-BREAKER");
            string currentState = circuitBreaker.ReturnCurrentData(CircuitBreaker.State);

            // Inbound-Request Processing
            if (context.IsRequest)
            {
                if (!String.IsNullOrEmpty(circuitBreaker.ReturnCurrentData(CircuitBreaker.OpenTimer)))
                {
                    circuitBreaker.ConfigureResponse();
                    context.TraceWriter.WriteMessage("Circuit Breaker is Open. Incoming requests will be blocked.");
                    return MessagePipelineResult.Reply;
                }

                if (currentState == CircuitBreaker.Open)
                {
                    circuitBreaker.SwitchToState(CircuitBreaker.HalfOpen);
                }
            }

            // Inbound-Response Processing
            else
            {
                HttpResponseMessageProperty responseMessage = context.Message.GetHttpResponseMessageProperty(false);
                int responseStatusCode = (int)responseMessage.StatusCode;

                if (circuitBreaker.IsExpectedResponse(responseStatusCode))
                {
                    if (currentState == CircuitBreaker.HalfOpen)
                    {
                        circuitBreaker.SwitchToState(CircuitBreaker.Closed);
                        context.TraceWriter.WriteMessage("Half-Open Circuit Breaker did not detect an error, switching to Closed state.");
                    }
                }
                else
                {
                    if (currentState == CircuitBreaker.Closed || String.IsNullOrEmpty(currentState))
                    {
                        circuitBreaker.IncreaseErrorCounter();

                        if (circuitBreaker.IsErrorLimitReached())
                        {
                            circuitBreaker.SwitchToState(CircuitBreaker.Open);
                            context.TraceWriter.WriteMessage("Error Limit of {0} is reached, switching to Open state.", circuitBreaker.ErrorLimit);
                        }
                    }
                    else
                    {
                        circuitBreaker.SwitchToState(CircuitBreaker.Open);
                        context.TraceWriter.WriteMessage("Half-Open Circuit Breaker detected an error, switching to Open state.");
                    }
                }
                
                int errorCounter = circuitBreaker.ReturnCurrentCounterData();

                if (errorCounter > 0)
                {
                    context.TraceWriter.WriteMessage("Current Errors: {0}", errorCounter);
                }
            }

            context.TraceWriter.WriteEnd(traceMark);
            return MessagePipelineResult.Continue;
        }
    }
}
